"""Annotate condition
"""
from typing import List, Any


from ..i18n import _
from ..error import DictConsistencyError
from ..config import Config

FREEZE_AUTOFREEZE_VARIABLE = 'module_instancie'


class ConditionAnnotator:
    """Annotate condition
    """
    def __init__(self,
                 objectspace,
                 ):
        self.objectspace = objectspace
        if hasattr(objectspace.space, 'variables'):
            self.convert_auto_freeze()
        if not hasattr(objectspace.space, 'constraints') or \
                not hasattr(self.objectspace.space.constraints, 'condition'):
            return
        self.convert_condition_target()
        self.convert_xxxlist_to_variable()
        self.check_condition_fallback()
        self.convert_condition_source()
        self.check_choice_option_condition()
        self.remove_condition_with_empty_target()
        self.convert_condition()

    def convert_auto_freeze(self):
        """convert auto_freeze
        only if FREEZE_AUTOFREEZE_VARIABLE == 'oui' this variable is frozen
        """
        def _convert_auto_freeze(variable):
            if not variable.auto_freeze:
                return
            if variable.namespace != Config['variable_namespace']:
                xmlfiles = self.objectspace.display_xmlfiles(variable.xmlfiles)
                msg = _(f'auto_freeze is not allowed in extra "{variable.namespace}" in {xmlfiles}')
                raise DictConsistencyError(msg, 49)
            new_condition = self.objectspace.condition(variable.xmlfiles)
            new_condition.name = 'auto_frozen_if_not_in'
            new_condition.namespace = variable.namespace
            new_condition.source = FREEZE_AUTOFREEZE_VARIABLE
            new_param = self.objectspace.param(variable.xmlfiles)
            new_param.text = 'oui'
            new_condition.param = [new_param]
            new_target = self.objectspace.target(variable.xmlfiles)
            new_target.type = 'variable'
            new_target.name = variable.name
            new_condition.target = [new_target]
            if not hasattr(self.objectspace.space, 'constraints'):
                self.objectspace.space.constraints = self.objectspace.constraints(variable.xmlfiles)
            if not hasattr(self.objectspace.space.constraints, 'condition'):
                self.objectspace.space.constraints.condition = []
            self.objectspace.space.constraints.condition.append(new_condition)
        for variables in self.objectspace.space.variables.values():
            for family in variables.family.values():
                if not hasattr(family, 'variable'):
                    continue
                for variable in family.variable.values():
                    if isinstance(variable, self.objectspace.leadership):
                        for follower in variable.variable:
                            _convert_auto_freeze(follower)
                    else:
                        _convert_auto_freeze(variable)

    def convert_condition_target(self):
        """verify and manage target in condition
        """
        for condition in self.objectspace.space.constraints.condition:
            if not hasattr(condition, 'target'):
                xmlfiles = self.objectspace.display_xmlfiles(condition.xmlfiles)
                msg = _(f'target is mandatory in a condition for source "{condition.source}" '
                        f'in {xmlfiles}')
                raise DictConsistencyError(msg, 9)
            remove_targets = []
            for index, target in enumerate(condition.target):
                try:
                    if target.type == 'variable':
                        if condition.source == target.name:
                            msg = _('target name and source name must be different: '
                                    f'{condition.source}')
                            raise DictConsistencyError(msg, 11)
                        target.name = self.objectspace.paths.get_variable(target.name)
                    elif target.type == 'family':
                        target.name = self.objectspace.paths.get_family(target.name,
                                                                        condition.namespace,
                                                                        )
                    elif target.type.endswith('list') and \
                            condition.name not in ['disabled_if_in', 'disabled_if_not_in']:
                        xmlfiles = self.objectspace.display_xmlfiles(target.xmlfiles)
                        msg = _(f'target "{target.type}" not allow in condition "{condition.name}" '
                                f'in {xmlfiles}')
                        raise DictConsistencyError(msg, 10)
                except DictConsistencyError as err:
                    if err.errno != 42:
                        raise err
                    # for optional variable
                    if not target.optional:
                        xmlfiles = self.objectspace.display_xmlfiles(condition.xmlfiles)
                        msg = f'cannot found target "{target.name}" in the condition in {xmlfiles}'
                        raise DictConsistencyError(_(msg), 12) from err
                    remove_targets.append(index)
            remove_targets.sort(reverse=True)
            for index in remove_targets:
                condition.target.pop(index)


    def convert_xxxlist_to_variable(self):
        """transform *list to variable or family
        """
        for condition in self.objectspace.space.constraints.condition:
            new_targets = []
            remove_targets = []
            for target_idx, target in enumerate(condition.target):
                if target.type.endswith('list'):
                    listname = target.type
                    listvars = self.objectspace.list_conditions.get(listname,
                                                                    {}).get(target.name)
                    if listvars:
                        for listvar in listvars:
                            type_ = 'variable'
                            new_target = self.objectspace.target(listvar.xmlfiles)
                            new_target.type = type_
                            new_target.name = listvar
                            new_targets.append(new_target)
                    remove_targets.append(target_idx)
            remove_targets.sort(reverse=True)
            for target_idx in remove_targets:
                condition.target.pop(target_idx)
            condition.target.extend(new_targets)

    def check_condition_fallback(self):
        """a condition with a fallback **and** the source variable doesn't exist
        """
        remove_conditions = []
        for idx, condition in enumerate(self.objectspace.space.constraints.condition):
            # fallback
            if condition.fallback is True and \
                    not self.objectspace.paths.path_is_defined(condition.source):
                apply_action = False
                if condition.name in ['disabled_if_in', 'mandatory_if_in', 'hidden_if_in']:
                    apply_action = not condition.force_condition_on_fallback
                else:
                    apply_action = condition.force_inverse_condition_on_fallback
                remove_conditions.append(idx)
                if apply_action:
                    self.force_actions_to_variable(condition)
        remove_conditions = list(set(remove_conditions))
        remove_conditions.sort(reverse=True)
        for idx in remove_conditions:
            self.objectspace.space.constraints.condition.pop(idx)

    def force_actions_to_variable(self,
                                  condition: 'self.objectspace.condition',
                                  ) -> None:
        """force property to a variable
        for example disabled_if_not_in => variable.disabled = True
        """
        actions = self.get_actions_from_condition(condition.name)
        for target in condition.target:
            leader_or_var, variables = self._get_family_variables_from_target(target)
            main_action = actions[0]
            setattr(leader_or_var, main_action, True)
            for action in actions[1:]:
                for variable in variables:
                    setattr(variable, action, True)

    @staticmethod
    def get_actions_from_condition(condition_name: str) -> List[str]:
        """get action's name from a condition
        """
        if condition_name.startswith('hidden_if_'):
            return ['hidden', 'frozen', 'force_default_on_freeze']
        if condition_name == 'auto_frozen_if_not_in':
            return ['auto_frozen']
        return [condition_name.split('_', 1)[0]]

    def _get_family_variables_from_target(self,
                                          target,
                                          ):
        if target.type == 'variable':
            if not self.objectspace.paths.is_leader(target.name.path):
                return target.name, [target.name]
            # it's a leader, so apply property to leadership
            family_name = self.objectspace.paths.get_variable_family_path(target.name.path)
            family = self.objectspace.paths.get_family(family_name,
                                                       target.name.namespace,
                                                       )
            return family, family.variable
        # it's a family
        variable = self.objectspace.paths.get_family(target.name.path,
                                                     target.namespace,
                                                     )
        return variable, list(variable.variable.values())

    def convert_condition_source(self):
        """remove condition for ChoiceOption that don't have param
        """
        for condition in self.objectspace.space.constraints.condition:
            try:
                condition.source = self.objectspace.paths.get_variable(condition.source)
            except DictConsistencyError as err:
                if err.errno == 36:
                    xmlfiles = self.objectspace.display_xmlfiles(condition.xmlfiles)
                    msg = _(f'the source "{condition.source}" in condition cannot be a dynamic '
                            f'variable in {xmlfiles}')
                    raise DictConsistencyError(msg, 20) from err

    def check_choice_option_condition(self):
        """remove condition for ChoiceOption that don't have param
        """
        remove_conditions = []
        for condition_idx, condition in enumerate(self.objectspace.space.constraints.condition):
            # FIXME only string?
            if condition.source.path in self.objectspace.valid_enums and \
                    self.objectspace.valid_enums[condition.source.path]['type'] == 'string':
                valid_enum = self.objectspace.valid_enums[condition.source.path]['values']
                remove_param = [param_idx for param_idx, param in enumerate(condition.param) \
                                if param.text not in valid_enum]
                remove_param.sort(reverse=True)
                for idx in remove_param:
                    del condition.param[idx]
                if not condition.param and condition.name.endswith('_if_not_in'):
                    self.force_actions_to_variable(condition)
                    remove_conditions.append(condition_idx)
        remove_conditions.sort(reverse=True)
        for idx in remove_conditions:
            self.objectspace.space.constraints.condition.pop(idx)

    def remove_condition_with_empty_target(self):
        """remove condition with empty target
        """
        # optional target are remove, condition could be empty
        remove_conditions = [condition_idx for condition_idx, condition in \
                             enumerate(self.objectspace.space.constraints.condition) \
                             if not condition.target]
        remove_conditions.sort(reverse=True)
        for idx in remove_conditions:
            self.objectspace.space.constraints.condition.pop(idx)

    def convert_condition(self):
        """valid and manage <condition>
        """
        for condition in self.objectspace.space.constraints.condition:
            actions = self.get_actions_from_condition(condition.name)
            for param in condition.param:
                text = getattr(param, 'text', None)
                for target in condition.target:
                    leader_or_variable, variables = self._get_family_variables_from_target(target)
                    # if option is already disable, do not apply disable_if_in
                    # check only the first action (example of multiple actions:
                    # 'hidden', 'frozen', 'force_default_on_freeze')
                    main_action = actions[0]
                    if getattr(leader_or_variable, main_action, False) is True:
                        continue
                    self.build_property(leader_or_variable,
                                        text,
                                        condition,
                                        main_action,
                                        )
                    if isinstance(leader_or_variable, self.objectspace.variable) and \
                            (leader_or_variable.auto_save or leader_or_variable.auto_freeze) and \
                            'force_default_on_freeze' in actions:
                        continue
                    for action in actions[1:]:
                        # other actions are set to the variable or children of family
                        for variable in variables:
                            self.build_property(variable,
                                                text,
                                                condition,
                                                action,
                                                )

    def build_property(self,
                       obj,
                       text: Any,
                       condition: 'self.objectspace.condition',
                       action: str,
                       ) -> 'self.objectspace.property_':
        """build property_ for a condition
        """
        prop = self.objectspace.property_(obj.xmlfiles)
        prop.type = 'calculation'
        prop.inverse = condition.name.endswith('_if_not_in')
        prop.source = condition.source
        prop.expected = text
        prop.name = action
        if not hasattr(obj, 'properties'):
            obj.properties = []
        obj.properties.append(prop)
