"""loader
flattened XML specific
"""
from .config import Config
from .annotator import CONVERT_OPTION
from .objspace import RootRougailObject


FUNC_TO_DICT = []


class Root():  # pylint: disable=R0903
    """Root classes
    """
    path = '.'


class BaseElt:  # pylint: disable=R0903
    """Base element
    """
    name = 'baseoption'
    doc = 'baseoption'
    path = '.'


class TiramisuReflector:
    """Convert object to tiramisu representation
    """
    def __init__(self,
                 objectspace,
                 funcs_path,
                 ):
        self.index = 0
        self.text = ["from importlib.machinery import SourceFileLoader",
                     "from importlib.util import spec_from_loader, module_from_spec",
                     f"loader = SourceFileLoader('func', '{funcs_path}')",
                     "spec = spec_from_loader(loader.name, loader)",
                     "func = module_from_spec(spec)",
                     "loader.exec_module(func)",
                     "for key, value in dict(locals()).items():",
                     "    if key != ['SourceFileLoader', 'func']:",
                     "        setattr(func, key, value)",
                     "try:",
                     "    from tiramisu3 import *",
                     "except:",
                     "    from tiramisu import *",
                     "from rougail.tiramisu import ConvertDynOptionDescription",
                     ]
        self.objectspace = objectspace
        self.make_tiramisu_objects()

    def make_tiramisu_objects(self) -> None:
        """make tiramisu objects
        """
        baseelt = BaseElt()
        self.set_name(baseelt)
        basefamily = Family(baseelt,
                            self.text,
                            False,
                            )
        for elt in self.reorder_family():
            self.populate_family(basefamily,
                                 elt,
                                 )
        # parse object
        baseelt.reflector_object.get()  # pylint: disable=E1101

    def reorder_family(self):
        """variable_namespace family has to be loaded before any other family
        because `extra` family could use `variable_namespace` variables.
        """
        if hasattr(self.objectspace.space, 'variables'):
            if Config['variable_namespace'] in self.objectspace.space.variables:
                yield self.objectspace.space.variables[Config['variable_namespace']]
            for elt, value in self.objectspace.space.variables.items():
                if elt != Config['variable_namespace']:
                    yield value
        if hasattr(self.objectspace.space, 'services'):
            yield self.objectspace.space.services

    def populate_family(self,
                        parent_family,
                        elt,
                        ):
        """Populate family
        """
        self.set_name(elt)
        family = Family(elt,
                        self.text,
                        isinstance(elt, self.objectspace.leadership),
                        )
        parent_family.add(family)
        for children in vars(elt).values():
            if isinstance(children, self.objectspace.family):
                self.populate_family(family,
                                     children,
                                     )
                continue
            if isinstance(children, dict):
                children = list(children.values())
            if isinstance(children, list):
                for child in children:
                    if isinstance(child, self.objectspace.property_) or \
                            not isinstance(child, RootRougailObject):
                        continue
                    if isinstance(child, self.objectspace.variable):
                        self.set_name(child)
                        family.add(Variable(child,
                                            self.text,
                                            self.objectspace,
                                            ))
                    else:
                        self.populate_family(family,
                                             child,
                                             )

    def set_name(self,
                 elt,
                 ):
        """Set name
        """
        elt.reflector_name = f'option_{self.index}'
        self.index += 1

    def get_text(self):
        """Get text
        """
        return '\n'.join(self.text)


class Common:
    """Common function for variable and family
    """
    def __init__(self,
                 elt,
                 text,
                 ):
        self.elt = elt
        self.option_name = None
        self.text = text
        self.elt.reflector_object = self
        self.object_type = None

    def get(self):
        """Get tiramisu's object
        """
        if self.option_name is None:
            self.option_name = self.elt.reflector_name
            self.populate_attrib()
            self.populate_informations()
        return self.option_name

    def populate_attrib(self):
        """Populate attributes
        """
        keys = {}
        keys['name'] = self.convert_str(self.elt.name)
        if hasattr(self.elt, 'doc'):
            keys['doc'] = self.convert_str(self.elt.doc)
        self._populate_attrib(keys)
        if hasattr(self.elt, 'properties'):
            keys['properties'] = self.properties_to_string(self.elt.properties)
        ret_list = []
        for key, value in keys.items():
            ret_list.append(f'{key}={value}')
        attrib = ', '.join(ret_list)
        self.text.append(f'{self.option_name} = {self.object_type}({attrib})')

    def _populate_attrib(self,
                         keys: dict,
                         ) -> None:
        raise NotImplementedError()

    @staticmethod
    def convert_str(value):
        """convert string
        """
        return "'" + value.replace("'", "\\\'") + "'"

    def properties_to_string(self,
                             value: list,
                             ) -> None:
        """Change properties to string
        """
        properties = []
        calc_properties = []
        for property_ in value:
            if not isinstance(property_, str):
                calc_properties.append(self.populate_properties(property_))
            else:
                properties.append(f"'{property_}'")
        return 'frozenset({' + ', '.join(sorted(properties) + calc_properties) + '})'

    @staticmethod
    def populate_properties(child) -> str:
        """Populate properties
        """
        assert child.type == 'calculation'
        action = f"ParamValue('{child.name}')"
        option_name = child.source.reflector_object.get()
        kwargs = (f"'condition': ParamOption({option_name}, todict=True), "
                  f"'expected': ParamValue('{child.expected}')")
        if child.inverse:
            kwargs += ", 'reverse_condition': ParamValue(True)"
        return 'Calculation(calc_value, Params(' + action + ', kwargs={' + kwargs + '}))'

    def populate_informations(self):
        """Populate Tiramisu's informations
        """
        if not hasattr(self.elt, 'information'):
            return
        for key, value in vars(self.elt.information).items():
            if key == 'xmlfiles':
                continue
            if isinstance(value, str):
                value = '"' + value.replace('"', '\"') + '"'
            self.text.append(f'{self.option_name}.impl_set_information("{key}", {value})')


class Variable(Common):
    """Manage variable
    """
    def __init__(self,
                 elt,
                 text,
                 objectspace,
                 ):
        super().__init__(elt, text)
        self.objectspace = objectspace
        convert_option = CONVERT_OPTION[elt.type]
        self.object_type = convert_option['opttype']

    def _populate_attrib(self,
                         keys: dict,
                         ):
        if hasattr(self.elt, 'opt'):
            keys['opt'] = self.elt.opt.reflector_object.get()
        if hasattr(self.elt, 'values'):
            values = self.elt.values
            if values[0].type == 'calculation':
                values = values[0].name.reflector_object.get()
                keys['values'] = f"Calculation(func.calc_value, Params((ParamOption({values}))))"
            else:
                keys['values'] = str(tuple([val.name for val in values]))
        if hasattr(self.elt, 'multi') and self.elt.multi:
            keys['multi'] = self.elt.multi
        for key in ['default', 'default_multi']:
            if hasattr(self.elt, key) and getattr(self.elt, key) is not None:
                value = getattr(self.elt, key)
                if isinstance(value, str):
                    value = self.convert_str(value)
                elif isinstance(value, self.objectspace.value):
                    value = self.calculation_value(value, [])
                keys[key] = value
        if hasattr(self.elt, 'validators'):
            keys['validators'] = '[' + ', '.join([self.calculation_value(val,
                                    ['ParamSelfOption()']) for val in self.elt.validators]) + ']'
        for key in ['min_number', 'max_number']:
            if hasattr(self.elt, key):
                keys[key] = getattr(self.elt, key)
        for key, value in CONVERT_OPTION[self.elt.type].get('initkwargs', {}).items():
            keys[key] = value

    def calculation_value(self,
                          child,
                          args,
                          ) -> str:
        """Generate calculated value
        """
        kwargs = []
        # has parameters
        function = child.name
        if hasattr(child, 'param'):
            for param in child.param:
                value = self.populate_param(function, param)
                if not hasattr(param, 'name'):
                    args.append(str(value))
                else:
                    kwargs.append(f"'{param.name}': " + value)
        ret = f"Calculation(func.{function}, Params((" + ', '.join(args) + \
                "), kwargs=" + "{" + ', '.join(kwargs) + "})"
        if hasattr(child, 'warnings_only'):
            ret += f', warnings_only={child.warnings_only}'
        return ret + ')'

    def populate_param(self,
                       function: str,
                       param,
                       ):
        """Populate variable parameters
        """
        if param.type == 'string':
            return f'ParamValue("{param.text}")'
        if param.type == 'number':
            return f'ParamValue({param.text})'
        if param.type == 'variable':
            value = {'option': param.text,
                     'notraisepropertyerror': param.notraisepropertyerror,
                     'todict': function in FUNC_TO_DICT,
                     }
            if hasattr(param, 'suffix'):
                value['suffix'] = param.suffix
                value['family'] = param.family
            return self.build_param(value)
        if param.type == 'information':
            return f'ParamInformation("{param.text}", None)'
        if param.type == 'suffix':
            return 'ParamSuffix()'
        return ''  # pragma: no cover

    @staticmethod
    def build_param(param) -> str:
        """build variable parameters
        """
        option_name = param['option'].reflector_object.get()
        ends = f"notraisepropertyerror={param['notraisepropertyerror']}, todict={param['todict']})"
        if 'suffix' in param:
            family_name = param['family'].reflector_name
            return f"ParamDynOption({option_name}, '{param['suffix']}', {family_name}, {ends}"
        return f"ParamOption({option_name}, {ends}"


class Family(Common):
    """Manage family
    """
    def __init__(self,
                 elt,
                 text,
                 is_leadership
                 ):
        super().__init__(elt, text)
        self.is_leadership = is_leadership
        if hasattr(self.elt, 'suffixes'):
            self.object_type = 'ConvertDynOptionDescription'
        elif self.is_leadership:
            self.object_type = 'Leadership'
        else:
            self.object_type = 'OptionDescription'
        self.children = []

    def add(self, child):
        """Add a child
        """
        self.children.append(child)

    def _populate_attrib(self,
                         keys: list,
                         ) -> None:
        if hasattr(self.elt, 'suffixes'):
            dyn = self.elt.suffixes.reflector_object.get()
            keys['suffixes'] = f"Calculation(func.calc_value, Params((ParamOption({dyn}))))"
        keys['children'] = '[' + ', '.join([child.get() for child in self.children]) + ']'
